/*
 * Basic I2C functions
 *
 * Copyright (c) 2004 Texas Instruments
 *
 * Copyright (c) 2007
 * Diego Santa Cruz, SpinetiX S.A., Diego.SantaCruz@spinetix.com
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Jian Zhang jzhang@ti.com, Texas Instruments
 *
 * Copyright (c) 2003 Wolfgang Denk, wd@denx.de
 * Rewritten to fit into the current U-Boot framework
 *
 * Adapted for DaVinci I2C, swami.iyer@ti.com
 *
 */

#include <common.h>

#ifdef CONFIG_DRIVER_DAVINCI_I2C

#include "davinci_i2c.h"
#include <i2c.h>
#include <asm/io.h>

#define inw(a) __raw_readw(a)
#define outw(v,a) __raw_writew(v,a)

static int wait_for_bb (void);
void flush_fifo(void);

#if 0
static int debug = 0;
#define INITDEBUG do { debug = (getenv("debug") != NULL); } while (0)
#define DEBUGF(x...) do { if (debug) printf(x); } while(0)
#else
#define INITDEBUG do { } while (0)
#define DEBUGF(x...) do { } while (0)
#endif

void i2c_init (int speed, int slaveadd)
{
	u16 scl;

	if (inw (I2C_CON) & I2C_CON_EN) {
		outw (0, I2C_CON);
		udelay (50000);
	}

	outw (2, I2C_PSC); /* sets clock to 9 MHz */
	scl = ( 9000000 + 2*speed - 1 ) / ( 2*speed );
	if (scl <= 5) { /* offset of 5 for SCLL and SCLH */
	     scl = 0;
	} else {
	     scl -= 5;
	}
	DEBUGF("scl = %u\n", scl);
	outw (scl, I2C_SCLL);
	outw (scl, I2C_SCLH);
	/* own address */
	outw (slaveadd, I2C_OA);
	outw (0, I2C_CNT);
	/* have to enable intrrupts or DaVinci i2c module doesn't work */
	outw (I2C_IE_SCD_IE | I2C_IE_XRDY_IE | I2C_IE_RRDY_IE | I2C_IE_ARDY_IE |
	      I2C_IE_NACK_IE | I2C_IE_AL_IE, I2C_IE);
	outw (I2C_CON_EN, I2C_CON);
	udelay (1000);

}

static int i2c_read_bytes (u8 devaddr, u8 values[], int len, int wait_bus)
{
	int i2c_error = 0;
	const int timeout = 100000;
	int timecount, i;
	u16 status;

	DEBUGF("i2c_read_bytes(): # bytes %d\n", len);

	/* wait until bus not busy */
	if (wait_bus && wait_for_bb ())
	     return 1;

	timecount = 0;
	while (!i2c_error) {
	     /* len bytes */
	     DEBUGF("count = %d\n", len);
	     outw (len, I2C_CNT);
	     /* set slave address */
	     outw (devaddr, I2C_SA);
	     /* need stop bit here */
	     outw (I2C_CON_EN | I2C_CON_MST | I2C_CON_STT | I2C_CON_STP,
		   I2C_CON);
	     udelay(10000);
	     timecount += 10000;
	     if ( (status = inw(I2C_STAT)) & I2C_STAT_NACK) { /* chip not responding */
		  outw (I2C_CON_EN, I2C_CON);
		  outw(0xFFFF, I2C_STAT);
		  outw (0, I2C_CNT);
		  if (!wait_bus || timecount > timeout) {
		       printf("I2C timeout in device select\n");
		       i2c_error = 1;
		  } else {
			  DEBUGF("chip did not respond to device select, trying again\n");
		  }
	     } else if (status & I2C_STAT_AL) {
		  printf("I2C arbitration lost\n");
		  i2c_error = 1;
	     } else {
		  break; /* chip responded */
	     }
	}

	for (i=0; i < len && !i2c_error; i++) {
		timecount = 0;
		while( ( (status = inw(I2C_STAT)) & I2C_STAT_RRDY) == 0 ) {
		     udelay(50);
		     timecount += 50;
		     if (timecount >= timeout) {
			 printf("I2C timeout reading value %i of %i\n", i+1, len);
			  i2c_error = 1;
			  break;
		     }
		}
		DEBUGF("status 0x%X, timecount %d\n", status, timecount);
		if (!i2c_error) {
		     values[i] = inw (I2C_DRR);
		     DEBUGF("got val 0x%X\n", values[i]);
		}
	}

	if (!i2c_error) {
	     do {
		  status = inw(I2C_STAT) & I2C_STAT_SCD;
	     } while (!status);	
	     outw (I2C_CON_EN, I2C_CON);
	}
	flush_fifo();
	outw (0xFFFF, I2C_STAT);
	outw (0, I2C_CNT);
	/* make sure we leave the bus free the minimum time after stop condition */
	udelay(CFG_I2C_WAIT_USEC);
	return i2c_error;
}

static int i2c_write_bytes (u8 devaddr, u8 regoffset, int alen, u8 values[], int len)
{
	int i2c_error = 0;
	const int timeout = 100000;
	int timecount, i;
	u16 status, val;

	DEBUGF("i2c_write_bytes(): addr %d, addr len %u, # bytes %d\n", regoffset, alen, len);

	/* wait until bus not busy */
	if (wait_for_bb ())
	     return 1;

	timecount = 0;
	while (!i2c_error) {
	     /* len+alen bytes */
	     DEBUGF("count = %d\n", len+alen);
	     outw (len+alen, I2C_CNT);
	     /* set slave address */
	     outw (devaddr, I2C_SA);
	     /* stop bit needed here if len > 0 */
	     outw (I2C_CON_EN | I2C_CON_MST | I2C_CON_STT | I2C_CON_TRX |
		   (len > 0 ? I2C_CON_STP : 0), I2C_CON);
	     udelay(10000);
	     timecount += 10000;
	     if ( (status = inw(I2C_STAT)) & I2C_STAT_NACK) { /* chip not responding */
		  outw (I2C_CON_EN, I2C_CON);
		  outw(0xFFFF, I2C_STAT);
		  outw (0, I2C_CNT);
		  if (timecount > timeout) {
		       printf("I2C timeout in device select\n");
		       i2c_error = 1;
		  } else {
			  DEBUGF("chip did not respond to device select, trying again\n");
		  }
	     } else if (status & I2C_STAT_AL) {
		  printf("I2C arbitration lost\n");
		  i2c_error = 1;
	     } else {
		  break; /* chip responded */
	     }
	}

	for (i=0; i < len+alen && !i2c_error; i++) {
	     val = (i < alen ? regoffset : values[i-alen]);
	     DEBUGF("sending val 0x%X\n", val);
	     outw(val, I2C_DXR);
	     timecount = 0;
	     while( ( (status = inw(I2C_STAT)) & I2C_STAT_XRDY) == 0 ) {
		  udelay(50);
		  timecount += 50;
		  if (timecount >= timeout) {
		      printf("I2C timeout writing byte %i of %i\n", i+1, len+alen);
		       i2c_error = 1;
		       break;
		  }
	     }
	     DEBUGF("status 0x%X, timecount %d\n", status, timecount);
	     if (status & I2C_STAT_NACK) {
		  i2c_error = 1;
	     }
	}

	if (!i2c_error) {
	     if (len > 0) {
		  do {
		       status = inw(I2C_STAT) & I2C_STAT_SCD;
		  } while (!status);	
	     }
	     outw (I2C_CON_EN, I2C_CON);
	}
	outw (0xFFFF, I2C_STAT);
	outw (0, I2C_CNT);
	/* make sure we leave the bus free the minimum time after stop condition */
	if (len > 0)
		udelay(CFG_I2C_WAIT_USEC);
	return i2c_error;
}

void flush_fifo(void)
{	u16 stat;

	/* note: if you try and read data when its not there or ready
	 * you get a bus error
	 */
	while(1){
		stat = inw(I2C_STAT);
		if(stat == I2C_STAT_RRDY){
			inw(I2C_DRR);
			outw(I2C_STAT_RRDY,I2C_STAT);
			udelay(1000);
		}else
			break;
	}
}

int i2c_probe (uchar chip)
{
	int res = 1; /* default = fail */

	if (chip == inw (I2C_OA)) {
		return res;
	}

	/* wait until bus not busy */
	if (wait_for_bb ())
	     return res;

	/* try to read one byte */
	outw (1, I2C_CNT);
	/* set slave address */
	outw (chip, I2C_SA);
	/* stop bit needed here */
	outw (I2C_CON_EN | I2C_CON_MST | I2C_CON_STT | I2C_CON_STP, I2C_CON);
	/* enough delay for the NACK bit set */
	udelay (50000);

	if (!(inw (I2C_STAT) & I2C_STAT_NACK)) {
		res = 0;      /* success case */
		flush_fifo();
		outw(0xFFFF, I2C_STAT);
	} else {
		outw(0xFFFF, I2C_STAT);	 /* failue, clear sources*/
		outw (inw (I2C_CON) | I2C_CON_STP, I2C_CON); /* finish up xfer */
		udelay(20000);
		wait_for_bb ();
	}
	flush_fifo();
	outw (0, I2C_CNT); /* don't allow any more data in...we don't want it.*/
	outw(0xFFFF, I2C_STAT);
	return res;
}

int i2c_read (uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	if (alen < 0 || alen > 1) {
		printf ("I2C read: addr len %d not supported\n", alen);
		return 1;
	}

	if (alen == 0)
	     addr = 0;

	if (len <= 0) {
		printf("I2C read: cannot read zero bytes\n");
		return 1;
	}

	if (addr + len > 256) {
		printf ("I2C read: address out of range\n");
		return 1;
	}

	INITDEBUG;

	if (alen) {
	     if (i2c_write_bytes(chip, addr, alen, NULL, 0)) {
		  printf("I2C read: I/O error while setting byte address\n");
		  i2c_init (CFG_I2C_SPEED, CFG_I2C_SLAVE);
		  return 1;
	     }
	     udelay(100); /* wait for device to process */
	}
	if (i2c_read_bytes(chip, buffer, len, !alen)) {
	     printf ("I2C read: I/O error\n");
	     i2c_init (CFG_I2C_SPEED, CFG_I2C_SLAVE);
	     return 1;
	}

	return 0;
}

int i2c_write (uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	if (alen < 0 || alen > 1) {
		printf ("I2C read: addr len %d not supported\n", alen);
		return 1;
	}

	if (alen == 0)
	     addr = 0;

	if (len <= 0) {
		printf("I2C write: cannot write zero bytes\n");
		return 1;
	}

	if (addr + len > 256) {
		printf ("I2C read: address out of range\n");
		return 1;
	}

	INITDEBUG;

	if (i2c_write_bytes (chip, addr, alen, buffer, len)) {
	     printf ("I2C write: I/O error\n");
	     i2c_init (CFG_I2C_SPEED, CFG_I2C_SLAVE);
	     return 1;
	}

	return 0;
}


uchar i2c_reg_read(uchar i2c_addr, uchar reg)
{
        char buf;

        i2c_read(i2c_addr, reg, 1, &buf, 1);

        return(buf);
}

void i2c_reg_write(uchar i2c_addr, uchar reg, uchar val)
{
        i2c_write(i2c_addr, reg, 1, &val, 1);
}

static int wait_for_bb (void)
{
	int timeout = 2;
	u16 stat;

	outw(0xFFFF, I2C_STAT);	 /* clear current interrupts...*/
	while ((stat = inw (I2C_STAT) & I2C_STAT_BB) && timeout--) {
		outw (stat, I2C_STAT);
		udelay (50000);
	}

	if (timeout <= 0) {
		printf ("I2C timed out waiting on busy bus (I2C_STAT=0x%X)\n",
			inw (I2C_STAT));
		return 1;
	}
	outw(0xFFFF, I2C_STAT);	 /* clear delayed stuff*/
	return 0;
}

#endif /* CONFIG_DRIVER_DAVINCI_I2C */
